import logging

class Log(object):
    class __Log:
        def __init__(self):
            self.val = None

            FORMAT = '%(message)s %(asctime)-15s %(clientip)s %(user)-8s'
            logging.basicConfig(format=FORMAT)            
                        
            self.logger = logging.getLogger('tcpserver')            
            self.d = {'clientip': "IP", 'user': "USER"}
            
        def __str__(self):
            return repr(self) + self.val
        
        def debugMode(self, flag):
            if flag:
                self.logger.setLevel(logging.DEBUG)
            else:
                self.logger.setLevel(logging.INFO)
        
        def info(self, msg):
            self.logger.setLevel(logging.INFO)
            self.logger.info('[INFO]: %s\t at', msg, extra=self.d)
            
        def warn(self, msg):            
            self.logger.setLevel(logging.WARNING)
            self.logger.warning('[WARN]: %s\t at', msg, extra=self.d)
            
        def error(self, msg):
            self.logger.setLevel(logging.ERROR)
            self.logger.error('[ERRO]: %s\t at', msg, extra=self.d)
        
        def debug(self, msg):            
            self.logger.debug('[DEBG]: %s\t at', msg, extra=self.d)
            
    instance = None
    def __new__(cls):
        if not Log.instance:
            Log.instance = Log.__Log()
        return Log.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name):
        return setattr(self.instance, name)
    


class HashTable:
    def __init__(self):
        self.size = 11
        self.slots = [None] * self.size
        self.data = [None] * self.size

    def put(self,key,data):
      hashvalue = self.hashfunction(key,len(self.slots))

      if self.slots[hashvalue] == None:
        self.slots[hashvalue] = key
        self.data[hashvalue] = data
      else:
        if self.slots[hashvalue] == key:
          self.data[hashvalue] = data  #replace
        else:
          nextslot = self.rehash(hashvalue,len(self.slots))
          while self.slots[nextslot] != None and \
                          self.slots[nextslot] != key:
            nextslot = self.rehash(nextslot,len(self.slots))

          if self.slots[nextslot] == None:
            self.slots[nextslot]=key
            self.data[nextslot]=data
          else:
            self.data[nextslot] = data #replace

    def hashfunction(self,key,size):
         return self.hasht(key,size)#key%size

    def hasht(self,astring, tablesize):
         sum = 0
         for pos in range(len(astring)):
            sum = sum + ord(astring[pos])
         return sum%tablesize

    def rehash(self,oldhash,size):
        return (oldhash+1)%size

    def get(self,key):
      startslot = self.hashfunction(key,len(self.slots))

      data = None
      stop = False
      found = False
      position = startslot
      while self.slots[position] != None and  \
                           not found and not stop:
         if self.slots[position] == key:
           found = True
           data = self.data[position]
         else:
           position=self.rehash(position,len(self.slots))
           if position == startslot:
               stop = True
      return data

    def __getitem__(self,key):
        return self.get(key)

    def __setitem__(self,key,data):
        self.put(key,data)



from threading import Thread
import time
import random
import traceback

class BehaviorExe(Thread):

    let = None
    alive = None
    queue = None

    def __init__(self, queue):
        self.let = False
        self.alive = True
        self.queue = queue
        super().__init__()

    def run(self):
        while self.alive:
            if self.let:
                evt = self.queue.get()
                self.queue.task_done()
                try:
                    evt['action'].execute(evt['data'])
                except Exception as e:
                    print(e)
                    traceback.print_exc()                

    def setLet(self, val):
        self.let = val

    def setAlive(self, val):
        self.alive = val



from .GameAdapter import GameAdapter
from .MongoAdapter import MongoAdapter
from .RESTAdapter import RESTAdapter
from .SubProcessAdapter import SubProcessAdapter
from .WSSAdapter import WSSAdapter
from .WSSNJAdapter import WSSNJAdapter
from .WSSNJHandler import WSSNJHandler



from abc import ABC, abstractmethod
from ...kernel.system.Adm import Adm
import logging

class Action(ABC):
    
    id = None
    adm = None
    agent = None
    log = None
    
    def __init__(self):        
        super().__init__()
        
    @abstractmethod
    def execute(self, data):
        pass

    def setAgent(self, agent):
    	self.adm = Adm()
    	self.agent = agent

    def setupLogger(self, name, log_file, logg):
        level = None
        if logg == 'debug': 
            level = logging.DEBUG
        if logg == 'info': 
            level = logging.INFO,
        if logg == 'warning': 
            level = logging.WARNING,
        if logg == 'error': 
            level = logging.ERROR,
        if logg == 'critical': 
            level = logging.CRITICAL
        self.log = logging.getLogger(name)
        if not self.log:        
            formatter = logging.Formatter('%(asctime)s %(name)-12s %(lineno)d %(levelname)-8s %(message)s')
            handler = logging.FileHandler(log_file)        
            handler.setFormatter(formatter)
            self.log = logging.getLogger(name)
            self.log.setLevel(level)
            self.log.addHandler(handler)
    
    def setupLocalLogger(self):
        formatter = logging.Formatter('%(asctime)s %(name)-12s %(lineno)d %(levelname)-8s %(message)s')
        handler = logging.FileHandler('agents.log')        
        handler.setFormatter(formatter)
        self.log = logging.getLogger('agents')
        self.log.setLevel(logging.INFO)
        self.log.addHandler(handler)



from abc import ABC, abstractmethod
from ...kernel.util.Queue import Queue
from ...kernel.system.Adm import Adm
from ...kernel.agent.Channel import Channel
from ...kernel.util.HashTable import HashTable
from ...kernel.agent.BehaviorExe import BehaviorExe

class Agent(ABC):
    
    id = None
    state = None
    settings = None    
    behaviors = None
    channelsTable = None
    behaviorsTable = None
    workerList = None
    channelList = None
            
    def __init__(self, arg):

        self.settings = self.setUp(arg)
        self.id = self.settings['id']
        self.state = self.settings['state']
        Adm().addAgent(self)

        self.eventsTable = HashTable()
        self.channelsTable = HashTable()

        self.workerList = []
        self.channelList = []

        self.behaviors = self.settings['behaviors']
        
        for beh in self.behaviors:            
            queue = Queue(10)
            channel = Channel(queue)    
            worker = BehaviorExe(queue)
            
            self.channelsTable[beh['name']] = {'channel' : channel, 'worker': worker}  

            self.workerList.append(worker)
            self.channelList.append(channel)

            events = beh['events']
            for evts in events:
                evts['action'].setAgent(self)
                self.eventsTable[evts['performative']] = {'behavior' : beh['name'], 'action': evts['action']} 

        super().__init__()
        
    @abstractmethod
    def setUp(self, settings):
        pass
    
    def sendEvent(self, event, data):
        behavior = self.eventsTable[event]
        channel = self.channelsTable[behavior['behavior']]
        evt = {'event': event, 'data': data, 'action': behavior['action']}              
        channel['channel'].sendEvent(evt)
    
    def start(self):
        for w in self.workerList:
            w.setLet(True)
            w.start()
            
    def wait(self):
        for w in self.workerList:
            w.setLet(False)

    def finalize(self):
        for w in self.workerList:
            w.setAlive(False)
    
    def kill(self):
        # TODO Call garbachcollector
        pass

    def toDTO(self):
        dto = {
            'command': 'MOVE',
            'class': self.__class__.__name__,
            'path': self.__module__,
            'id': self.id,
            'state': self.state  
        }
        rtn = str(dto)
        rtn = rtn.replace("'", "\"")  
        return rtn

    def getState(self):
        return self.state



# -*- coding: utf-8 -*-
import os
import sys
import json
import socket
from time import sleep
from ...kernel.util.Log import Log
from ...kernel.util.HashTable import HashTable
from ...kernel.system.Directory import Directory
from ...kernel.adapter.FileAdapter import FileAdapter
from ...middleware.remote.RemoteAdm import RemoteAdm
from ...middleware.remote.AdmListener import AdmListener

class Adm(object):
    class __Adm:
        def __init__(self):
            self.val = None
            self.conf = None
            self.adapters = HashTable()
            self.agentsTable = HashTable()
            self.containerList = []
            
        def __str__(self):
            return repr(self) + self.val

        def start(self):
            self.startByConf('')

        def startByConf(self, conf):            
            log = Log()

            if conf == '':
                self.conf = {
                    "user" : "local",
                    "ip" : "localhost",
                    "port" : 8080,
                    "remote" : None     
                }
            else:
                CONF_DIR = conf
                fa = FileAdapter({'alias':'JsonAdapter', 'type': 'JSON', 'path': CONF_DIR})
                fa.setUp()
                param = fa.request()
                self.conf = param['conf']

            if self.conf['remote']:
                remote = self.conf['remote']                
                if remote['mode'] == 'MASTER':
                    self.remoteAdm = RemoteAdm(remote['ip'], remote['port'])
                    self.remoteAdm.start()    
                else:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)                        
                    attempts = remote['attempts']
                    master = remote['master']
                    data = '{"command": "REGISTER", "alias":"' + remote['container_alias'] + '", "ip":"' + remote['ip'] + '", "port":"'+ str(remote['port']) +'"}'
                    for x in range(1, attempts):                        
                        try:                            
                            sock.connect((master['master_ip'], master['master_port']))                            
                            sock.sendall(bytes(data + "\n", "utf-8"))
                            received = str(sock.recv(1024), "utf-8")
                            print("Received: {}".format(received))                                                            
                            if received == 'ACK' + "\n":
                                break
                        except:
                            log.warn('The container could not connect to the master. Attempt:' + str(x))
                            sleep(1) # 1 second.
                        finally:
                            sock.close()

                    admListener = AdmListener(remote['ip'], remote['port'])
                    admListener.start()
            
        def addAgent(self, agent):
            self.agentsTable[agent.id] = agent
            ip = ''
            port = ''
            if self.conf['remote']:
                remote = self.conf['remote'] 
                ip = remote['ip']
                port = remote['port']
            directory = Directory()
            directory.addAgent({'agent': agent.id, 'ip': ip, 'port' : port})
            if self.conf['remote']:
                remote = self.conf['remote']
                agents = directory.getAgents()
                dto = '{"command":"UPDATE", "agents" : ' + json.dumps(agents, ensure_ascii=False) + '}'                
                if remote['mode'] == 'MASTER':
                    containers = directory.getContainers()                    
                    for ctn in containers:
                        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        try:                            
                            sock.connect( (ctn['ip'], int(ctn['port'])) )                            
                            sock.sendall(bytes(dto + "\n", "utf-8"))
                        except:
                            log.warn('The container could not connect to container:' + ctn['alias'])
                        finally:
                            sock.close()
                else:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)                        
                    attempts = remote['attempts']
                    master = remote['master']                                              
                    try:                            
                        sock.connect((master['master_ip'], master['master_port']))                            
                        sock.sendall(bytes(dto + "\n", "utf-8"))
                    except:
                        log.warn('The container could not connect to the master.')
                        sleep(1) # 1 second.
                    finally:
                        sock.close()
            
        def sendEvent(self, agentID, event, data):
            try:
                ag = self.agentsTable[agentID]
                if (ag):
                    ag.sendEvent(event, data)
                    return True
                else:
                    log = Log()
                    remote = self.conf['remote'] 
                    attempts = remote['attempts']
                    agents = Directory().getAgents()
                    for agent in agents:
                        if agentID == agent['agent']:
                            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                            aux = str(data)
                            aux = aux.replace("\'", "\"")                                                
                            dto = '{"command": "SENDEVENT", "alias":"' + agentID + '", "event":"' + event + '", "data":'+ aux +'}'
                            print('Send::')
                            print(dto)                               
                            for x in range(1, attempts):                        
                                try:                            
                                    sock.connect((agent['ip'], agent['port']))                            
                                    sock.sendall(bytes(dto + "\n", "utf-8"))
                                    received = str(sock.recv(1024), "utf-8")
                                    print("Received: {}".format(received))                                                            
                                    if received == 'ACK' + "\n":
                                        break
                                except:
                                    log.warn('The container could not connect to the master. Attempt:' + str(x))
                                    sleep(1) # 1 second.
                                finally:
                                    sock.close()
                    return True

            except Exception as inst:
                print('Controled ecxeption in Adm: ')
                print(inst)
                return False

        def addAdapter(self, id, adapter):
            self.adapters[id] = adapter

        def getAdapter(self, id):
            return self.adapters[id]

        def getConf(self):
            return self.conf

        def addContainer(self, container):
            self.containerList.append(container)

        def moveAgent(self, agentID, container):
            ag = self.agentsTable[agentID]
            ag.finalize()
            dto = ag.toDTO()
            print (dto)
            containers = Directory().getContainers()
            for ctn in containers:
                if container == ctn['alias']:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    try:                            
                        sock.connect( (ctn['ip'], int(ctn['port'])) )                            
                        sock.sendall(bytes(dto + "\n", "utf-8"))
                    except:
                        Log().warn('The container could not connect to container: ' + ctn['alias'])
                    finally:
                        sock.close()
                    break    
            self.agentsTable[agentID] = None
            Directory().removeAgent(agentID)

        def setEndPoint(self, agID, socket):
            ag = self.agentsTable[agID]
            ag.setSocket(socket)
        
    instance = None
    def __new__(cls):
        if not Adm.instance:
            Adm.instance = Adm.__Adm()
        return Adm.instance
    def __getattr__(self, name):
        return getattr(self.instance, name)
    def __setattr__(self, name):
        return setattr(self.instance, name)
